import json
import time

import httpx
from cachetools import TTLCache, cached

from .algorithms import get_algorithm
from .core import b64url_decode, b64url_encode
from .exceptions import (
    JWKNotFoundError,
    JWTExpiredError,
    JWTInvalidFormatError,
    JWTInvalidSignatureError,
    JWTIssuedInFutureError,
    JWTNotValidYetError,
)


def extract_jwt_parts(token: str) -> tuple[dict, dict, bytes, bytes]:
    try:
        header_b64, payload_b64, signature_b64 = token.split(".")
        header = json.loads(b64url_decode(header_b64))
        payload = json.loads(b64url_decode(payload_b64))
        signature = b64url_decode(signature_b64)
        signing_input = f"{header_b64}.{payload_b64}".encode()
        return header, payload, signature, signing_input
    except (ValueError, json.JSONDecodeError) as e:
        raise JWTInvalidFormatError(f"Invalid JWT format: {str(e)}")


@cached(cache=TTLCache(maxsize=1000, ttl=3600))
def fetch_jwk(*, jwks_url: str, kid: str) -> dict | None:
    jwks = httpx.get(jwks_url).json()
    for key in jwks["keys"]:
        if key["kid"] == kid:
            return key
    raise JWKNotFoundError(f"JWK with kid '{kid}' not found")


def verify_temporal_claims(*, payload: dict):
    now = int(time.time())
    if "exp" in payload and now >= payload["exp"]:
        raise JWTExpiredError("Token expired")
    if "nbf" in payload and now < payload["nbf"]:
        raise JWTNotValidYetError("Token not valid yet (nbf)")
    if "iat" in payload and now < payload["iat"] - 60:
        raise JWTIssuedInFutureError("Token issued in the future (iat)")
    return True


def verify_signature(
    *, alg: str, key: dict, data: bytes | str | dict, signature: bytes
) -> bool:
    """
    Verify a JWT signature using the specified algorithm and key.

    Args:
        alg: The algorithm used for signing
        key: The verification key (JWK dict)
        signing_input: The data that was signed
        signature: The signature to verify

    Returns:
        bool: True if the signature is valid, False otherwise

    Raises:
        JWTInvalidSignatureError: If the signature verification fails
    """
    if isinstance(data, dict):
        signing_input = b64url_encode(json.dumps(data)).encode()
    elif isinstance(data, str):
        signing_input = data.encode()
    else:
        signing_input = data

    algorithm = get_algorithm(alg)
    if not algorithm.verify(data=signing_input, signature=signature, key=key, alg=alg):
        raise JWTInvalidSignatureError("Invalid signature")
    return True


def verify_jwt(
    *, token: str, jwks_url: str = None, kid: str = None, jwk: dict | None = None
) -> bool:
    header, payload, signature, signing_input = extract_jwt_parts(token)
    if jwk is None:
        jwk = fetch_jwk(jwks_url=jwks_url, kid=kid)

    if not verify_signature(
        alg=header["alg"], key=jwk, data=signing_input, signature=signature
    ):
        raise JWTInvalidSignatureError("Invalid signature")

    return verify_temporal_claims(payload=payload)



"""JWT-related exceptions."""


class JWTError(Exception):
    """Base exception for JWT-related errors."""


class JWTExpiredError(JWTError):
    """Raised when a JWT has expired."""


class JWTNotValidYetError(JWTError):
    """Raised when a JWT's 'nbf' claim indicates it's not valid yet."""


class JWTIssuedInFutureError(JWTError):
    """Raised when a JWT's 'iat' claim is in the future."""


class JWTInvalidSignatureError(JWTError):
    """Raised when a JWT's signature is invalid."""


class JWTInvalidFormatError(JWTError):
    """Raised when a JWT has an invalid format."""


class JWKNotFoundError(JWTError):
    """Raised when a JWK with the specified kid is not found."""



from typing import TypeVar

from cachetools import TTLCache, cached
from pydantic import BaseModel

from .core import Algorithm
from .verify import (
    extract_jwt_parts,
    fetch_jwk,
    verify_signature,
    verify_temporal_claims,
)

T = TypeVar("T", bound="BaseModel")


class JWT(BaseModel):
    token: str
    key: dict | None = None
    jwks_url: str | None = None

    def __init__(
        self,
        *,
        token: str,
        key: dict | None = None,
        jwks_url: str | None = None,
        payload_class: type[T] | None = None,
    ):
        super().__init__(token=token, key=key, jwks_url=jwks_url)
        self._payload_class = payload_class

    def __hash__(self) -> int:
        return hash(self.token)

    @property
    @cached(cache=TTLCache(maxsize=1000, ttl=300))
    def _parts(self) -> tuple[dict, dict, bytes, bytes]:
        return extract_jwt_parts(self.token)

    @property
    def unverified_header(self) -> dict[str, str]:
        return self._parts[0]

    @property
    def header(self) -> dict[str, str]:
        if self.verify():
            return self.unverified_header
        raise ValueError("JWT is not valid")

    @property
    def algorithm(self) -> Algorithm:
        return Algorithm(self.unverified_header["alg"].upper())

    @property
    def unverified_payload(self) -> dict | T:
        if self._payload_class is not None:
            return self._payload_class.model_validate(self._parts[1])
        return self._parts[1]
    
    @property
    def payload(self) -> dict | T:
        if self.verify():
            return self.unverified_payload
        raise ValueError("JWT is not valid")

    @property
    def signature(self) -> bytes:
        return self._parts[2]

    @property
    def signing_input(self) -> bytes:
        return self._parts[3]

    def verify(self) -> bool:
        if self.jwks_url is None and self.key is None:
            raise ValueError("Either jwks_url or key must be provided")

        if self.jwks_url is not None:
            self.key = fetch_jwk(self.unverified_header["kid"], self.jwks_url)

        if self.key is None:
            raise ValueError("Key must be provided")

        verification = verify_signature(
            alg=self.algorithm,
            key=self.key,
            data=self.signing_input,
            signature=self.signature,
        )
        if not verification:
            return False
        return verify_temporal_claims(payload=self.unverified_payload)
